/*
 * Copyright 2008 Perforce Software.  All rights reserved.
 *
 * This file is part of Perforce - the FAST SCM System.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "MyDTS.h"
#include "MyDTG.h"
extern "C" {
#include <dtg-utils.h>
}
#include <dtg-str.h>

const char *MyDTGProj::MyDTGMagic = "MyDTGProjClass";

MyDTGProj::MyDTGProj( MyDTG *dt, const char *proj, struct DTGError *error  )
{
	magic = MyDTGMagic;
	in_dt = dt;
	seg_filters = NULL;
	testing = dt->testing;
	clear_DTGError( error );
	if( !dt || !proj )
	{
	    set_DTGError( error, "MyDTGProj::MyDTGProj: Undefined arguments" );
	    return;
	}
	project = mk_string( proj );
	char *err = NULL;
	testing = dt->testing;
	if ( !testing )
		in_dt->dts->connect_to_project( project, err );
	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = in_dt->dts->is_valid();
	    return;
	}
}

MyDTGProj::~MyDTGProj()
{
	if( project )
	    delete[] project;
	if( seg_filters )
	    delete[] seg_filters;
}

// When updating field-processing here, keep extract_filter_string() in sync.
struct DTGFieldDesc *MyDTGProj::list_fields( struct DTGError *error )
{
	struct DTGFieldDesc *list;
	clear_DTGError( error );
	if( testing ) 
	{
	    list = new_DTGFieldDesc( "JobID", "word", 1, NULL );
	    return list;
	}

	char *err = NULL;
	list = in_dt->dts->get_field_desc( project, err );
	if( !list || err )
	{
	    set_DTGError( error, err );
	    error->can_continue = in_dt->dts->is_valid();
	    delete[] err;
	    delete_DTGFieldDesc( list );
	    return NULL;
	}

	return list;
}

struct DTGStrList *MyDTGProj::list_changed_defects( int max_rows,
						  struct DTGDate *since, 
	                                          const char *mod_date_field,
	                                          const char *mod_by_field,
	                                          const char *exclude_user,
	                                          struct DTGError *error )
{
	struct DTGStrList *list;
	if( testing ) 
	{
	    list = new_DTGStrList( "*defect*" );
	    clear_DTGError( error );
	    return list;
	}
	char *err = NULL;
	list = in_dt->dts->list_jobs( max_rows, 
					since, mod_date_field,
					exclude_user, mod_by_field, 
					seg_filters, err );
	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = in_dt->dts->is_valid();
	    delete[] err;
	    delete_DTGStrList( list );
	    return NULL;
	}

	return list;
}

MyDTGDefect *MyDTGProj::get_defect( const char *defect, struct DTGError *error )
{
	MyDTGDefect *item = new MyDTGDefect( this, defect, error );
	if( error->message )
	{
	    delete item;
	    item = NULL;
	}
	return item;
}

MyDTGDefect *MyDTGProj::new_defect( struct DTGError *error )
{
	const char *err = mk_string( "The ", in_dt->get_name( error ),
	                             " plugin does ",
	                             "not support creating new defects." );
	set_DTGError( error, err );
	delete[] err;
	return NULL;
}

/*
	Takes the list_fields() output and decomposes it into an SQL query to be
	appended to what MyDTS::list_jobs() uses.

	It's safe to only query these fields once and not worry about new fields
	being created in Bugzilla since DTG segmentation is based off of
	pre-existing fields.

	The field names passed in are not Bugzilla database table names, so they
	must be converted.  E.g. 'OS/Version' -> `op_sys`
*/
static char *extract_filter_string( MyDTG *in_dt, struct DTGFieldDesc *f )
{
	struct DTGField *field_names = in_dt->dts->field_names();
	char *query = NULL, *tmp = NULL;

	for( ; f; f = f->next )
	{
	    if( !f->select_values )
	    {
	        delete[] query;
	        delete_DTGField( field_names );
	        return NULL;
	    }

	    // DTG pseudo-fields are no-ops.  E.g. DTGConfig-User
	    if( !strncmp( "DTG", f->name, 3 ) )
	        continue;

	    // E.g. f->name = 'Severity', f->select_values = trivial,enhancement
	    for( struct DTGStrList *i = f->select_values; i; i = i->next )
	    {
	        char *tmp1 = NULL;

	        // These fields are more free-form, so we escape them.
	        char *esc = in_dt->dts->esc_field( i->value );
	        tmp1 = tmp;
	        tmp = mk_string( tmp, "`", in_dt->dts->find_value(
	                         field_names, f->name ), "` = '", esc, "'");
	        delete[] tmp1;
	        free( esc );

	        if( i->next )
	        {
	            char *tmp2 = tmp;
	            tmp = mk_string( tmp, " OR " );
	            delete[] tmp2;
	        }
	    }

	    char *subquery = mk_string( "( ", tmp, " )" );
	    delete[] tmp;
	    tmp = mk_string( query, " AND ", subquery );
	    delete[] subquery;
	    delete[] query;
	    query = tmp;
	}
	delete_DTGField( field_names );
	return query;
}

void MyDTGProj::segment_filters( struct DTGFieldDesc *filters )
{
	delete[] seg_filters;
	seg_filters = extract_filter_string( in_dt, filters );
	if( in_dt->cur_message )
	    delete[] in_dt->cur_message;
	in_dt->cur_message = mk_string( "Segment filter is: ", seg_filters );
	in_dt->cur_message_level = 1;
}

